#!/usr/bin/env python

#
# Copyright 2021, Heidelberg University Clinic
#
# File author(s): Sebastian Lobentanzer
#                 ...
#
# Distributed under MIT licence, see the file `LICENSE`.
#
"""
Module data directory, including:

* The BioLink database schema
* The default config files
"""

from typing import Any, Optional
import os

import yaml
import appdirs

__all__ = ['module_data', 'module_data_path', 'read_config', 'config', 'reset']


def module_data_path(name: str) -> str:
    """
    Absolute path to a YAML file shipped with the module.
    """

    here = os.path.dirname(os.path.abspath(__file__))

    return os.path.join(here, f'{name}.yaml')


def module_data(name: str) -> Any:
    """
    Retrieve the contents of a YAML file shipped with this module.
    """

    path = module_data_path(name)

    return _read_yaml(path)


def _read_yaml(path: str) -> Optional[dict]:

    if os.path.exists(path):

        with open(path, 'r') as fp:

            return yaml.load(fp.read(), Loader=yaml.FullLoader)


def read_config() -> dict:
    """
    Read the module config.

    Read and merge the built-in default, the user level and directory level
    configuration.

    TODO explain path configuration
    """

    defaults = module_data('biocypher_config')
    user_confdir = appdirs.user_config_dir('biocypher', 'saezlab')
    user = _read_yaml(os.path.join(user_confdir, 'biocypher_config.yaml')) or {}
    # check if there is a local config file `biocypher_config.yaml` in the
    # current working directory # TODO account for .yml?
    if os.path.exists('biocypher_config.yaml'):
        local = _read_yaml('biocypher_config.yaml')
    elif os.path.exists('config/biocypher_config.yaml'):
        local = _read_yaml('config/biocypher_config.yaml')
    else:
        local = {}

    defaults.update(user)
    defaults.update(local)

    return defaults


def config(*args, **kwargs) -> Optional[Any]:
    """
    Set or get module config parameters.
    """

    if args and kwargs:

        raise ValueError(
            'Setting and getting values in the same call is not allowed.',
        )

    if args:

        result = tuple(globals()['_config'].get(key, None) for key in args)

        return result[0] if len(result) == 1 else result

    globals()['_config'].update(kwargs)


def reset():
    """
    Reload configuration from the config files.
    """

    globals()['_config'] = read_config()


reset()



#!/usr/bin/env python
#
# Copyright 2021, Heidelberg University Clinic
#
# File author(s): Sebastian Lobentanzer
#                 ...
#
# Distributed under MIT licence, see the file `LICENSE`.
#
"""
Create a property graph database for biomedical research applications.
Transforms ordered collections of biomedical entities and relationships
to BioCypher objects that represent property graph nodes and edges.

Todo:

    - Calls to the classes are independent, so there is no way to
      check directly; nodes can be created at any point in time
      previous to edge creation. We could require a pass of all
      the nodes in the graph when creating edges. Pro: this would
      also allow a check whether the existing graph adheres to
      BioCypher, at least in the node domain. If it doesn't,
      the call does not make much sense.

    - establish a dictionary lookup with the id types to be used / basic
      type checking of the input

    - translation of id types using pypath translation facilities (to be
      later externalised)
"""

from typing import Union, Optional
from datetime import datetime
from dataclasses import field, dataclass
from urllib.request import urlopen
import os

import yaml

from . import _misc
from . import _config as config
from ._logger import logger

__all__ = [
    'BioCypherEdge',
    'BioCypherNode',
    'BioCypherRelAsNode',
    'VersionNode',
]

logger.debug(f'Loading module {__name__}.')


@dataclass(frozen=True)
class BioCypherNode:
    """
    Handoff class to represent biomedical entities as Neo4j nodes.

    Has id, label, property dict; id and label (in the Neo4j sense of a
    label, ie, the entity descriptor after the colon, such as
    ":Protein") are non-optional and called node_id and node_label to
    avoid confusion with "label" properties. Node labels are written in
    PascalCase and as nouns, as per Neo4j consensus.

    Args:
        node_id (string): consensus "best" id for biological entity
        node_label (string): primary type of entity, capitalised
        **properties (kwargs): collection of all other properties to be
            passed to neo4j for the respective node (dict)

    Todo:
        - check and correct small inconsistencies such as capitalisation
            of ID names ("uniprot" vs "UniProt")
        - check for correct ID patterns (eg "ENSG" + string of numbers,
            uniprot length)
        - ID conversion using pypath translation facilities for now
    """

    node_id: str
    node_label: str
    preferred_id: str = 'id'
    properties: dict = field(default_factory=dict)

    def __post_init__(self):
        """
        Add id field to properties.

        Check for reserved keywords.

        Replace unwanted characters in properties.
        """
        self.properties['id'] = self.node_id
        self.properties['preferred_id'] = self.preferred_id or None
        # TODO actually make None possible here; as is, "id" is the default in
        # the dataclass as well as in the configuration file

        if ':TYPE' in self.properties.keys():
            logger.warning(
                "Keyword ':TYPE' is reserved for Neo4j. "
                'Removing from properties.',
                # "Renaming to 'type'."
            )
            # self.properties["type"] = self.properties[":TYPE"]
            del self.properties[':TYPE']

        for k, v in self.properties.items():
            if isinstance(v, str):
                self.properties[k] = (
                    v.replace(
                        os.linesep,
                        ' ',
                    ).replace(
                        '\n',
                        ' ',
                    ).replace(
                        '\r',
                        ' ',
                    ).replace('"', "'")
                )

            elif isinstance(v, list):
                self.properties[k] = (
                    [
                        val.replace(
                            os.linesep,
                            ' ',
                        ).replace(
                            '\n',
                            ' ',
                        ).replace('\r', ' ') for val in v
                    ]
                )

    def get_id(self) -> str:
        """
        Returns primary node identifier.

        Returns:
            str: node_id
        """
        return self.node_id

    def get_label(self) -> str:
        """
        Returns primary node label.

        Returns:
            str: node_label
        """
        return self.node_label

    def get_preferred_id(self) -> str:
        """
        Returns preferred id.

        Returns:
            str: preferred_id
        """
        return self.preferred_id

    def get_properties(self) -> dict:
        """
        Returns all other node properties apart from primary id and
        label as key-value pairs.

        Returns:
            dict: properties
        """
        return self.properties

    def get_dict(self) -> dict:
        """
        Return dict of id, labels, and properties.

        Returns:
            dict: node_id and node_label as top-level key-value pairs,
            properties as second-level dict.
        """
        return {
            'node_id': self.node_id,
            'node_label': self.node_label,
            'properties': self.properties,
        }


@dataclass(frozen=True)
class BioCypherEdge:
    """
    Handoff class to represent biomedical relationships in Neo4j.

    Has source and target ids, label, property dict; ids and label (in
    the Neo4j sense of a label, ie, the entity descriptor after the
    colon, such as ":TARGETS") are non-optional and called source_id,
    target_id, and relationship_label to avoid confusion with properties
    called "label", which usually denotes the human-readable form.
    Relationship labels are written in UPPERCASE and as verbs, as per
    Neo4j consensus.

    Args:

        source_id (string): consensus "best" id for biological entity

        target_id (string): consensus "best" id for biological entity

        relationship_label (string): type of interaction, UPPERCASE

        properties (dict): collection of all other properties of the
        respective edge

    """

    source_id: str
    target_id: str
    relationship_label: str
    relationship_id: str = None
    properties: dict = field(default_factory=dict)

    def __post_init__(self):
        """
        Check for reserved keywords.
        """

        if ':TYPE' in self.properties.keys():
            logger.debug(
                "Keyword ':TYPE' is reserved for Neo4j. "
                'Removing from properties.',
                # "Renaming to 'type'."
            )
            # self.properties["type"] = self.properties[":TYPE"]
            del self.properties[':TYPE']

    def get_id(self) -> Union[str, None]:
        """
        Returns primary node identifier or None.

        Returns:
            str: node_id
        """

        return self.relationship_id

    def get_source_id(self) -> str:
        """
        Returns primary node identifier of relationship source.

        Returns:
            str: source_id
        """
        return self.source_id

    def get_target_id(self) -> str:
        """
        Returns primary node identifier of relationship target.

        Returns:
            str: target_id
        """
        return self.target_id

    def get_label(self) -> str:
        """
        Returns relationship label.

        Returns:
            str: relationship_label
        """
        return self.relationship_label

    def get_properties(self) -> dict:
        """
        Returns all other relationship properties apart from primary ids
        and label as key-value pairs.

        Returns:
            dict: properties
        """
        return self.properties

    def get_dict(self) -> dict:
        """
        Return dict of ids, label, and properties.

        Returns:
            dict: source_id, target_id and relationship_label as
                top-level key-value pairs, properties as second-level
                dict.
        """
        return {
            'source_id': self.source_id,
            'target_id': self.target_id,
            'relationship_label': self.relationship_label,
            'properties': self.properties,
        }


@dataclass(frozen=True)
class BioCypherRelAsNode:
    """
    Class to represent relationships as nodes (with in- and outgoing
    edges) as a triplet of a BioCypherNode and two BioCypherEdges. Main
    usage in type checking (instances where the receiving function needs
    to check whether it receives a relationship as a single edge or as
    a triplet).

    Args:

        node (BioCypherNode): node representing the relationship

        source_edge (BioCypherEdge): edge representing the source of the
            relationship

        target_edge (BioCypherEdge): edge representing the target of the
            relationship

    """

    node: BioCypherNode
    source_edge: BioCypherEdge
    target_edge: BioCypherEdge

    def __post_init__(self):
        if not isinstance(self.node, BioCypherNode):
            raise TypeError(
                f'BioCypherRelAsNode.node must be a BioCypherNode, '
                f'not {type(self.node)}.',
            )

        if not isinstance(self.source_edge, BioCypherEdge):
            raise TypeError(
                f'BioCypherRelAsNode.source_edge must be a BioCypherEdge, '
                f'not {type(self.source_edge)}.',
            )

        if not isinstance(self.target_edge, BioCypherEdge):
            raise TypeError(
                f'BioCypherRelAsNode.target_edge must be a BioCypherEdge, '
                f'not {type(self.target_edge)}.',
            )

    def get_node(self):
        return self.node

    def get_source_edge(self):
        return self.source_edge

    def get_target_edge(self) -> BioCypherEdge:
        return self.target_edge


class VersionNode:
    """
    Versioning and graph structure information meta node. Similar to
    BioCypherNode but fixes label to ":BioCypher" and sets version by using the
    current date and time (meaning it overrides both mandatory args from
    BioCypherNode).

    Is created upon establishment of connection with the database and remains
    fixed for each BioCypher "session" (ie, the entire duration from starting
    the connection to the termination of the BioCypher adapter instance). Is
    connected to MetaNodes and MetaEdges via ":CONTAINS" relationships.
    """
    def __init__(
        self,
        offline: bool = False,
        from_config: bool = False,
        config_file: str = None,
        node_label: str = 'BioCypher',
        bcy_driver=None,
    ):

        logger.warning('VersionNode is deprecated.')

        # if we do not have a driver, then likely we are offline, right?
        self.offline = offline or getattr(bcy_driver, 'offline', True)
        self.from_config = from_config
        self.config_file = config_file
        self.node_label = node_label
        self.bcy_driver = bcy_driver

        self.node_id = self._get_current_id()
        self.graph_state = (
            self._get_graph_state() if not self.offline else None
        )
        self.schema = self._get_graph_schema()
        self.extended_schema = self._get_leaves()

        self.properties = {
            'graph_state': self.graph_state,
            'schema': self.schema,
            'leaves': self.extended_schema,
        }

    def get_id(self) -> str:
        """
        Returns primary node identifier.

        Returns:
            str: node_id
        """
        return self.node_id

    def get_label(self) -> str:
        """
        Returns primary node label.

        Returns:
            str: node_label
        """
        return self.node_label

    def get_dict(self) -> dict:
        """
        Return dict of id, labels, and properties.

        Returns:
            dict: node_id and node_label as top-level key-value pairs,
            properties as second-level dict.
        """
        return {
            'node_id': self.node_id,
            'node_label': self.node_label,
            'properties': self.properties,
        }

    def _get_current_id(self):
        """
        Instantiate a version ID for the current session. For now does
        versioning using datetime.

        Can later implement incremental versioning, versioning from
        config file, or manual specification via argument.
        """

        now = datetime.now()
        return now.strftime('v%Y%m%d-%H%M%S')

    def _get_graph_state(self):
        """
        Check in active DBMS connection for existence of VersionNodes,
        return the most recent VersionNode as representation of the
        graph state. If no VersionNode found, assume blank graph state
        and initialise.
        """

        logger.info('Getting graph state.')

        result, summary = self.bcy_driver.query(
            'MATCH (meta:BioCypher)'
            'WHERE NOT (meta)-[:PRECEDES]->(:BioCypher)'
            'RETURN meta',
        )

        # if result is empty, initialise
        if not result:
            logger.info('No existing graph found, initialising.')
            return None
        # else, pass on graph state
        else:
            version = result[0]['meta']['id']
            logger.info(f'Found graph state at {version}.')
            return result[0]['meta']

    def _get_graph_schema(
        self,
        from_config: Optional[bool] = None,
        config_file: Optional[str] = None,
    ) -> dict:
        """
        Return graph schema information from meta graph if it exists, or
        create new schema information properties from configuration
        file.

        Todo:
            - get schema from meta graph
        """

        from_config = self.from_config if from_config is None else from_config
        config_file = config_file or self.config_file

        if self.graph_state and not from_config:
            # TODO do we want information about actual structure here?
            res = self.bcy_driver.query(
                'MATCH (src:MetaNode) '
                # "OPTIONAL MATCH (src)-[r]->(tar)"
                'RETURN src',  # , type(r) AS type, tar"
            )
            gs_dict = {}
            for r in res[0]:
                src = r['src']
                key = src.pop('id')
                gs_dict[key] = src

            return gs_dict

        else:
            # load default yaml from module
            if config_file is None:

                dataMap = config.module_data('schema_config')

            # load yaml file from web
            elif config_file.startswith('http'):

                with urlopen(config_file) as f:

                    dataMap = yaml.safe_load(f)

            # get graph state from config (assume file is local)
            else:

                with open(config_file, 'r') as f:

                    dataMap = yaml.safe_load(f)

            return dataMap

    def _get_leaves(self, d: Optional[dict] = None) -> dict:
        """
        Get leaves of the tree hierarchy from the data structure dict
        contained in the `schema_config.yaml`. Creates virtual leaves
        (as children) from entries that provide more than one preferred
        id type (and corresponding inputs).

        Args:
            d:
                Data structure dict from yaml file.

        """

        d = d or self.schema

        leaves = dict()
        max_depth = 0  # TODO needed?

        # first pass: get parent leaves with direct representation in ontology
        for k, v in d.items():

            # k is not an entity
            if 'represented_as' not in v:
                continue

            # preferred_id optional: if not provided, use `id`
            if not v.get('preferred_id'):
                v['preferred_id'] = 'id'

            # k is an entity that is present in the ontology
            if 'is_a' not in v:
                leaves[k] = v

            # find max depth of children
            else:
                lst = v['is_a'] if isinstance(v['is_a'], list) else [v['is_a']]
                max_depth = max(max_depth, len(lst))

        # second pass: "vertical" inheritance
        d = self._vertical_property_inheritance(d)
        # create leaves for all straight descendants (no multiple identifiers
        # or sources) -> explicit children
        # TODO do we need to order children by depth from real leaves?
        leaves.update({k: v for k, v in d.items() if 'is_a' in v})

        # "horizontal" inheritance: create siblings for multiple identifiers or
        # sources -> virtual leaves or implicit children
        mi_leaves = {}
        ms_leaves = {}
        for k, v in d.items():

            # k is not an entity
            if 'represented_as' not in v:
                continue

            if isinstance(v.get('preferred_id'), list):
                mi_leaves = self._horizontal_inheritance_pid(k, v)
                leaves.update(mi_leaves)

            elif isinstance(v.get('source'), list):
                ms_leaves = self._horizontal_inheritance_source(k, v)
                leaves.update(ms_leaves)

        return leaves

    def _vertical_property_inheritance(self, d):
        """
        Inherit properties from parents to children and update `d` accordingly.
        """
        for k, v in d.items():

            # k is not an entity
            if 'represented_as' not in v:
                continue

            # k is an entity that is present in the ontology
            if 'is_a' not in v:
                continue

            # "vertical" inheritance: inherit properties from parent
            if v.get('inherit_properties', False):

                # get direct ancestor
                if isinstance(v['is_a'], list):
                    parent = v['is_a'][0]
                else:
                    parent = v['is_a']

                # update properties of child
                if self.schema[parent].get('properties'):
                    v['properties'] = self.schema[parent]['properties']
                if self.schema[parent].get('exclude_properties'):
                    v['exclude_properties'] = self.schema[parent][
                        'exclude_properties']

                # update schema (d)
                d[k] = v

        return d

    def _horizontal_inheritance_pid(self, key, value):
        """
        Create virtual leaves for multiple preferred id types or sources.

        If we create virtual leaves, label_in_input always has to be a list.
        """

        leaves = {}

        preferred_id = value['preferred_id']
        label_in_input = value['label_in_input']
        represented_as = value['represented_as']

        # adjust lengths
        max_l = max(
            [
                len(_misc.to_list(preferred_id)),
                len(_misc.to_list(label_in_input)),
                len(_misc.to_list(represented_as)),
            ],
        )

        # adjust pid length if necessary
        if isinstance(preferred_id, str):
            pids = [preferred_id] * max_l
        else:
            pids = preferred_id

        # adjust rep length if necessary
        if isinstance(represented_as, str):
            reps = [represented_as] * max_l
        else:
            reps = represented_as

        for pid, lab, rep in zip(pids, label_in_input, reps):

            skey = pid + '.' + key
            svalue = {
                'preferred_id': pid,
                'label_in_input': lab,
                'represented_as': rep,
                # mark as virtual
                'virtual': True,
            }

            # inherit is_a if exists
            if 'is_a' in value.keys():

                # treat as multiple inheritance
                if isinstance(value['is_a'], list):
                    v = list(value['is_a'])
                    v.insert(0, key)
                    svalue['is_a'] = v

                else:
                    svalue['is_a'] = [key, value['is_a']]

            else:
                # set parent as is_a
                svalue['is_a'] = key

            # inherit everything except core attributes
            for k, v in value.items():
                if k not in [
                    'is_a',
                    'preferred_id',
                    'label_in_input',
                    'represented_as',
                ]:
                    svalue[k] = v

            leaves[skey] = svalue

        return leaves

    def _horizontal_inheritance_source(self, key, value):
        """
        Create virtual leaves for multiple sources.

        If we create virtual leaves, label_in_input always has to be a list.
        """

        leaves = {}

        source = value['source']
        label_in_input = value['label_in_input']
        represented_as = value['represented_as']

        # adjust lengths
        src_l = len(source)

        # adjust label length if necessary
        if isinstance(label_in_input, str):
            labels = [label_in_input] * src_l
        else:
            labels = label_in_input

        # adjust rep length if necessary
        if isinstance(represented_as, str):
            reps = [represented_as] * src_l
        else:
            reps = represented_as

        for src, lab, rep in zip(source, labels, reps):

            skey = src + '.' + key
            svalue = {
                'source': src,
                'label_in_input': lab,
                'represented_as': rep,
                # mark as virtual
                'virtual': True,
            }

            # inherit is_a if exists
            if 'is_a' in value.keys():

                # treat as multiple inheritance
                if isinstance(value['is_a'], list):
                    v = list(value['is_a'])
                    v.insert(0, key)
                    svalue['is_a'] = v

                else:
                    svalue['is_a'] = [key, value['is_a']]

            else:
                # set parent as is_a
                svalue['is_a'] = key

            # inherit everything except core attributes
            for k, v in value.items():
                if k not in [
                    'is_a',
                    'source',
                    'label_in_input',
                    'represented_as',
                ]:
                    svalue[k] = v

            leaves[skey] = svalue

        return leaves



#!/usr/bin/env python

#
# Copyright 2021, Heidelberg University Clinic
#
# File author(s): Sebastian Lobentanzer
#                 ...
#
# Distributed under MIT licence, see the file `LICENSE`.
#

from typing import Optional
from urllib.request import urlopen

import yaml

from . import _misc
from ._config import config as _config
from ._logger import logger


class OntologyMapping:
    """
    Class to store the ontology mapping and extensions.
    """
    def __init__(self, config_file: str = None):

        self.schema = self._read_config(config_file)

        self.extended_schema = self._extend_schema()

    def _read_config(self, config_file: str = None):
        """
        Read the configuration file and store the ontology mapping and extensions.
        """
        if config_file is None:

            schema_config = _config.module_data('schema_config')

        # load yaml file from web
        elif config_file.startswith('http'):

            with urlopen(config_file) as f:

                schema_config = yaml.safe_load(f)

        # get graph state from config (assume file is local)
        else:

            with open(config_file, 'r') as f:

                schema_config = yaml.safe_load(f)

        return schema_config

    def _extend_schema(self, d: Optional[dict] = None) -> dict:
        """
        Get leaves of the tree hierarchy from the data structure dict
        contained in the `schema_config.yaml`. Creates virtual leaves
        (as children) from entries that provide more than one preferred
        id type (and corresponding inputs).

        Args:
            d:
                Data structure dict from yaml file.

        """

        d = d or self.schema

        extended_schema = dict()

        # first pass: get parent leaves with direct representation in ontology
        for k, v in d.items():

            # k is not an entity
            if 'represented_as' not in v:
                continue

            # preferred_id optional: if not provided, use `id`
            if not v.get('preferred_id'):
                v['preferred_id'] = 'id'

            # k is an entity that is present in the ontology
            if 'is_a' not in v:
                extended_schema[k] = v

        # second pass: "vertical" inheritance
        d = self._vertical_property_inheritance(d)
        for k, v in d.items():
            if 'is_a' in v:

                # prevent loops
                if k == v['is_a']:
                    logger.warning(
                        f'Loop detected in ontology mapping: {k} -> {v}. '
                        'Removing item. Please fix the inheritance if you want '
                        'to use this item.'
                    )
                    continue

                extended_schema[k] = v

        # "horizontal" inheritance: create siblings for multiple identifiers or
        # sources -> virtual leaves or implicit children
        mi_leaves = {}
        ms_leaves = {}
        for k, v in d.items():

            # k is not an entity
            if 'represented_as' not in v:
                continue

            if isinstance(v.get('preferred_id'), list):
                mi_leaves = self._horizontal_inheritance_pid(k, v)
                extended_schema.update(mi_leaves)

            elif isinstance(v.get('source'), list):
                ms_leaves = self._horizontal_inheritance_source(k, v)
                extended_schema.update(ms_leaves)

        return extended_schema

    def _vertical_property_inheritance(self, d):
        """
        Inherit properties from parents to children and update `d` accordingly.
        """
        for k, v in d.items():

            # k is not an entity
            if 'represented_as' not in v:
                continue

            # k is an entity that is present in the ontology
            if 'is_a' not in v:
                continue

            # "vertical" inheritance: inherit properties from parent
            if v.get('inherit_properties', False):

                # get direct ancestor
                if isinstance(v['is_a'], list):
                    parent = v['is_a'][0]
                else:
                    parent = v['is_a']

                # ensure child has properties and exclude_properties
                if 'properties' not in v:
                    v['properties'] = {}
                if 'exclude_properties' not in v:
                    v['exclude_properties'] = {}

                # update properties of child
                parent_props = self.schema[parent].get('properties', {})
                if parent_props:
                    v['properties'].update(parent_props)

                parent_excl_props = self.schema[parent].get(
                    'exclude_properties', {}
                )
                if parent_excl_props:
                    v['exclude_properties'].update(parent_excl_props)

                # update schema (d)
                d[k] = v

        return d

    def _horizontal_inheritance_pid(self, key, value):
        """
        Create virtual leaves for multiple preferred id types or sources.

        If we create virtual leaves, label_in_input always has to be a list.
        """

        leaves = {}

        preferred_id = value['preferred_id']
        label_in_input = value['label_in_input']
        represented_as = value['represented_as']

        # adjust lengths
        max_l = max(
            [
                len(_misc.to_list(preferred_id)),
                len(_misc.to_list(label_in_input)),
                len(_misc.to_list(represented_as)),
            ],
        )

        # adjust pid length if necessary
        if isinstance(preferred_id, str):
            pids = [preferred_id] * max_l
        else:
            pids = preferred_id

        # adjust rep length if necessary
        if isinstance(represented_as, str):
            reps = [represented_as] * max_l
        else:
            reps = represented_as

        for pid, lab, rep in zip(pids, label_in_input, reps):

            skey = pid + '.' + key
            svalue = {
                'preferred_id': pid,
                'label_in_input': lab,
                'represented_as': rep,
                # mark as virtual
                'virtual': True,
            }

            # inherit is_a if exists
            if 'is_a' in value.keys():

                # treat as multiple inheritance
                if isinstance(value['is_a'], list):
                    v = list(value['is_a'])
                    v.insert(0, key)
                    svalue['is_a'] = v

                else:
                    svalue['is_a'] = [key, value['is_a']]

            else:
                # set parent as is_a
                svalue['is_a'] = key

            # inherit everything except core attributes
            for k, v in value.items():
                if k not in [
                    'is_a',
                    'preferred_id',
                    'label_in_input',
                    'represented_as',
                ]:
                    svalue[k] = v

            leaves[skey] = svalue

        return leaves

    def _horizontal_inheritance_source(self, key, value):
        """
        Create virtual leaves for multiple sources.

        If we create virtual leaves, label_in_input always has to be a list.
        """

        leaves = {}

        source = value['source']
        label_in_input = value['label_in_input']
        represented_as = value['represented_as']

        # adjust lengths
        src_l = len(source)

        # adjust label length if necessary
        if isinstance(label_in_input, str):
            labels = [label_in_input] * src_l
        else:
            labels = label_in_input

        # adjust rep length if necessary
        if isinstance(represented_as, str):
            reps = [represented_as] * src_l
        else:
            reps = represented_as

        for src, lab, rep in zip(source, labels, reps):

            skey = src + '.' + key
            svalue = {
                'source': src,
                'label_in_input': lab,
                'represented_as': rep,
                # mark as virtual
                'virtual': True,
            }

            # inherit is_a if exists
            if 'is_a' in value.keys():

                # treat as multiple inheritance
                if isinstance(value['is_a'], list):
                    v = list(value['is_a'])
                    v.insert(0, key)
                    svalue['is_a'] = v

                else:
                    svalue['is_a'] = [key, value['is_a']]

            else:
                # set parent as is_a
                svalue['is_a'] = key

            # inherit everything except core attributes
            for k, v in value.items():
                if k not in [
                    'is_a',
                    'source',
                    'label_in_input',
                    'represented_as',
                ]:
                    svalue[k] = v

            leaves[skey] = svalue

        return leaves



#!/usr/bin/env python

#
# Copyright 2021, Heidelberg University Clinic
#
# File author(s): Sebastian Lobentanzer
#                 ...
#
# Distributed under MIT licence, see the file `LICENSE`.
#
"""
Export of CSV files for the Neo4J admin import. The admin import is able
to quickly transfer large amounts of content into an unused database. For more
explanation, see https://neo4j.com/docs/operations-manual/current/tuto\
rial/neo4j-admin-import/.

Import like that:
https://community.neo4j.com/t/how-can-i-use-a-database-created-with-neo4j-\
admin-import-in-neo4j-desktop/40594

    - Can properties the node/relationship does not own be left blank?

Formatting: --delimiter=";"
            --array-delimiter="|"
            --quote="'"

The header contains information for each field, for ID and properties
in the format <name>: <field_type>. E.g.:
`UniProtKB:ID;genesymbol;entrez_id:int;:LABEL`. Multiple labels can
be given by separating with the array delimiter.

There are three mandatory fields for relationship data:
:START_ID — ID referring to a node.
:END_ID — ID referring to a node.
:TYPE — The relationship type.

E.g.: `:START_ID;relationship_id;residue;:END_ID;:TYPE`.

Headers would best be separate files, data files with similar name but
different ending. Example from Neo4j documentation:

.. code-block:: bash

   bin/neo4j-admin import --database=neo4j
   --nodes=import/entities-header.csv,import/entities-part1.csv,
    import/entities-part2.csv
   --nodes=import/interactions-header.csv,import/interactions-part1.csv,
    import/interaction-part2.csv
   --relationships=import/rels-header.csv,import/rels-part1.csv,
    import/rels-part2.csv

Can use regex, e.g., [..] import/rels-part*. In this case, use padding
for ordering of the earlier part files ("01, 02").

# How to import:

1. stop the db

2. shell command:

.. code-block:: bash

   bin/neo4j-admin import --database=neo4j
   # nodes per type, separate header, regex for parts:
   --nodes="<path>/<node_type>-header.csv,<path>/<node_type>-part.*"
   # edges per type, separate header, regex for parts:
   --relationships="<path>/<edge_type>-header.csv,<path>/<edge_type>-part.*"

3. start db, test for consistency
"""

import glob

from ._logger import logger

logger.debug(f'Loading module {__name__}.')

from types import GeneratorType
from typing import TYPE_CHECKING, Union, Optional
from datetime import datetime
from collections import OrderedDict, defaultdict
import os

from more_itertools import peekable

from ._config import config as _config
from ._create import BioCypherEdge, BioCypherNode, BioCypherRelAsNode
from ._ontology import Ontology

__all__ = ['BatchWriter']

if TYPE_CHECKING:

    from ._translate import Translator, OntologyAdapter

# TODO retrospective check of written csvs?


class BatchWriter:
    """
    Class for writing node and edge representations to disk using the
    format specified by Neo4j for the use of admin import. Each batch
    writer instance has a fixed representation that needs to be passed
    at instantiation via the :py:attr:`schema` argument. The instance
    also expects an ontology adapter via :py:attr:`ontology_adapter` to be able
    to convert and extend the hierarchy.

    Args:
        ontology:
            Instance of :py:class:`Ontology` to enable translation and
            ontology queries

        translator:
            Instance of :py:class:`Translator` to enable translation of
            nodes and manipulation of properties.

        delimiter:
            The delimiter to use for the CSV files.

        array_delimiter:
            The delimiter to use for array properties.

        quote:
            The quote character to use for the CSV files.

        dirname:
            Path for exporting CSV files.

        db_name:
            Name of the Neo4j database that will be used in the generated
            commands.

        skip_bad_relationships:
            Whether to skip relationships that do not have a valid
            start and end node. (In admin import call.)

        skip_duplicate_nodes:
            Whether to skip duplicate nodes. (In admin import call.)

        import_call_bin_prefix:
            Path prefix for the admin import call binary.

        import_call_file_prefix:
            Path prefix for the data files (headers and parts) in the admin
            import call.

        wipe:
            Whether to force import (removing existing DB content). (In
            admin import call.)

        strict_mode:
            Whether to enforce source, version, and license properties.
    """
    def __init__(
        self,
        ontology: 'Ontology',
        translator: 'Translator',
        delimiter: str,
        array_delimiter: str,
        quote: str,
        dirname: Optional[str] = None,
        db_name: str = 'neo4j',
        skip_bad_relationships: bool = False,
        skip_duplicate_nodes: bool = False,
        import_call_bin_prefix: Optional[str] = None,
        import_call_file_prefix: Optional[str] = None,
        wipe: bool = True,
        strict_mode: bool = False,
    ):
        self.db_name = db_name

        self.delim = delimiter
        self.adelim = array_delimiter
        self.quote = quote
        self.skip_bad_relationships = skip_bad_relationships
        self.skip_duplicate_nodes = skip_duplicate_nodes

        if import_call_bin_prefix is None:
            self.import_call_bin_prefix = 'bin/'
        else:
            self.import_call_bin_prefix = import_call_bin_prefix

        self.wipe = wipe
        self.strict_mode = strict_mode

        self.extended_schema = ontology.extended_schema
        self.ontology = ontology
        self.translator = translator
        self.node_property_dict = {}
        self.edge_property_dict = {}
        self.import_call_nodes = ''
        self.import_call_edges = ''

        timestamp = lambda: datetime.now().strftime('%Y%m%d%H%M')

        self.outdir = dirname or os.path.join(_config('outdir'), timestamp())
        self.outdir = os.path.abspath(self.outdir)

        if import_call_file_prefix is None:
            self.import_call_file_prefix = self.outdir
        else:
            self.import_call_file_prefix = import_call_file_prefix

        logger.info(f'Creating output directory `{self.outdir}`.')
        os.makedirs(self.outdir, exist_ok=True)

        self.seen_node_ids = set()  # set to store the ids of nodes that have
        # already been written; to avoid duplicates
        self.duplicate_node_ids = set(
        )  # set to store the ids of nodes that were
        # found to have duplicates (avoid overloading the log)
        self.duplicate_node_types = set(
        )  # set to store the types of nodes that
        # have been found to have duplicates

        self.seen_edges = {}  # dict to store the set of edges that
        # have already been written; to avoid duplicates; per edge type
        self.duplicate_edge_ids = set()  # set to store the ids of edges that
        # were found to have duplicates (avoid overloading the log)
        self.duplicate_edge_types = set(
        )  # set to store the types of edges that
        # have been found to have duplicates

        # TODO not memory efficient, but should be fine for most cases; is
        # there a more elegant solution?

    def write_nodes(self, nodes, batch_size=int(1e6)):
        """
        Wrapper for writing nodes and their headers.

        Args:
            nodes (BioCypherNode): a list or generator of nodes in
                :py:class:`BioCypherNode` format

        Returns:
            bool: The return value. True for success, False otherwise.
        """
        # TODO check represented_as

        # write node data
        passed = self._write_node_data(nodes, batch_size)
        if not passed:
            logger.error('Error while writing node data.')
            return False
        # pass property data to header writer per node type written
        passed = self._write_node_headers()
        if not passed:
            logger.error('Error while writing node headers.')
            return False

        return True

    def write_edges(
        self,
        edges: Union[list, GeneratorType],
        batch_size: int = int(1e6),
    ) -> bool:
        """
        Wrapper for writing edges and their headers.

        Args:
            edges (BioCypherEdge): a list or generator of edges in
                :py:class:`BioCypherEdge` or :py:class:`BioCypherRelAsNode`
                format

        Returns:
            bool: The return value. True for success, False otherwise.
        """
        passed = False
        # unwrap generator in one step
        edges = list(edges)  # force evaluation to handle empty generator
        if edges:
            z = zip(
                *(
                    (
                        e.get_node(),
                        [
                            e.get_source_edge(),
                            e.get_target_edge(),
                        ],
                    ) if isinstance(e, BioCypherRelAsNode) else (None, [e])
                    for e in edges
                )
            )
            nod, edg = (list(a) for a in z)
            nod = [n for n in nod if n]
            edg = [val for sublist in edg for val in sublist]  # flatten

            if nod and edg:
                passed = self.write_nodes(nod) and self._write_edge_data(
                    edg,
                    batch_size,
                )
            else:
                passed = self._write_edge_data(edg, batch_size)

        else:
            # is this a problem? if the generator or list is empty, we
            # don't write anything.
            logger.debug(
                'No edges to write, possibly due to no matched Biolink classes.',
            )
            pass

        if not passed:
            logger.error('Error while writing edge data.')
            return False
        # pass property data to header writer per edge type written
        passed = self._write_edge_headers()
        if not passed:
            logger.error('Error while writing edge headers.')
            return False

        return True

    def _write_node_data(self, nodes, batch_size):
        """
        Writes biocypher nodes to CSV conforming to the headers created
        with `_write_node_headers()`, and is actually required to be run
        before calling `_write_node_headers()` to set the
        :py:attr:`self.node_property_dict` for passing the node properties
        to the instance. Expects list or generator of nodes from the
        :py:class:`BioCypherNode` class.

        Args:
            nodes (BioCypherNode): a list or generator of nodes in
                :py:class:`BioCypherNode` format

        Returns:
            bool: The return value. True for success, False otherwise.
        """

        if isinstance(nodes, GeneratorType) or isinstance(nodes, peekable):
            logger.debug('Writing node CSV from generator.')

            bins = defaultdict(list)  # dict to store a list for each
            # label that is passed in
            bin_l = {}  # dict to store the length of each list for
            # batching cutoff
            reference_props = defaultdict(
                dict,
            )  # dict to store a dict of properties
            # for each label to check for consistency and their type
            # for now, relevant for `int`
            labels = {}  # dict to store the additional labels for each
            # primary graph constituent from biolink hierarchy
            for node in nodes:
                _id = node.get_id()
                label = node.get_label()

                # check for non-id
                if not _id:
                    logger.warning(f'Node {label} has no id; skipping.')
                    continue

                # check if node has already been written, if so skip
                if _id in self.seen_node_ids:
                    self.duplicate_node_ids.add(_id)
                    if not label in self.duplicate_node_types:
                        self.duplicate_node_types.add(label)
                        logger.warning(
                            f'Duplicate nodes found in type {label}. '
                        )
                    continue

                if not label in bins.keys():
                    # start new list
                    all_labels = None
                    bins[label].append(node)
                    bin_l[label] = 1

                    # get properties from config if present
                    cprops = self.extended_schema.get(label).get('properties', )
                    if cprops:
                        d = dict(cprops)

                        # add id and preferred id to properties; these are
                        # created in node creation (`_create.BioCypherNode`)
                        d['id'] = 'str'
                        d['preferred_id'] = 'str'

                        # add strict mode properties
                        if self.strict_mode:
                            d['source'] = 'str'
                            d['version'] = 'str'
                            d['licence'] = 'str'

                    else:
                        d = dict(node.get_properties())
                        # encode property type
                        for k, v in d.items():
                            if d[k] is not None:
                                d[k] = type(v).__name__
                    # else use first encountered node to define properties for
                    # checking; could later be by checking all nodes but much
                    # more complicated, particularly involving batch writing
                    # (would require "do-overs"). for now, we output a warning
                    # if node properties diverge from reference properties (in
                    # write_single_node_list_to_file) TODO if it occurs, ask
                    # user to select desired properties and restart the process

                    reference_props[label] = d

                    # get label hierarchy
                    # multiple labels:
                    all_labels = self.ontology.get_ancestors(label)

                    if all_labels:
                        # convert to pascal case
                        all_labels = [
                            self.translator.name_sentence_to_pascal(label)
                            for label in all_labels
                        ]
                        # remove duplicates
                        all_labels = list(OrderedDict.fromkeys(all_labels))
                        # order alphabetically
                        all_labels.sort()
                        # concatenate with array delimiter
                        all_labels = self.adelim.join(all_labels)
                    else:
                        all_labels = self.translator.name_sentence_to_pascal(
                            label
                        )

                    labels[label] = all_labels

                else:
                    # add to list
                    bins[label].append(node)
                    bin_l[label] += 1
                    if not bin_l[label] < batch_size:
                        # batch size controlled here
                        passed = self._write_single_node_list_to_file(
                            bins[label],
                            label,
                            reference_props[label],
                            labels[label],
                        )

                        if not passed:
                            return False

                        bins[label] = []
                        bin_l[label] = 0

                self.seen_node_ids.add(_id)

            # after generator depleted, write remainder of bins
            for label, nl in bins.items():
                passed = self._write_single_node_list_to_file(
                    nl,
                    label,
                    reference_props[label],
                    labels[label],
                )

                if not passed:
                    return False

            # use complete bin list to write header files
            # TODO if a node type has varying properties
            # (ie missingness), we'd need to collect all possible
            # properties in the generator pass

            # save config or first-node properties to instance attribute
            for label in reference_props.keys():
                self.node_property_dict[label] = reference_props[label]

            return True
        else:
            if type(nodes) is not list:
                logger.error('Nodes must be passed as list or generator.')
                return False
            else:

                def gen(nodes):
                    yield from nodes

                return self._write_node_data(gen(nodes), batch_size=batch_size)

    def _write_node_headers(self):
        """
        Writes single CSV file for a graph entity that is represented
        as a node as per the definition in the `schema_config.yaml`,
        containing only the header for this type of node.

        Returns:
            bool: The return value. True for success, False otherwise.
        """
        # load headers from data parse
        if not self.node_property_dict:
            logger.error(
                'Header information not found. Was the data parsed first?',
            )
            return False

        for label, props in self.node_property_dict.items():
            # create header CSV with ID, properties, labels

            _id = ':ID'

            # to programmatically define properties to be written, the
            # data would have to be parsed before writing the header.
            # alternatively, desired properties can also be provided
            # via the schema_config.yaml.

            # translate label to PascalCase
            pascal_label = self.translator.name_sentence_to_pascal(label)

            header_path = os.path.join(
                self.outdir,
                f'{pascal_label}-header.csv',
            )
            parts_path = os.path.join(self.outdir, f'{pascal_label}-part.*')

            # check if file already exists
            if not os.path.exists(header_path):

                # concatenate key:value in props
                props_list = []
                for k, v in props.items():
                    if v in ['int', 'long']:
                        props_list.append(f'{k}:long')
                    elif v in ['float', 'double', 'dbl']:
                        props_list.append(f'{k}:double')
                    elif v in ['bool', 'boolean']:
                        # TODO Neo4j boolean support / spelling?
                        props_list.append(f'{k}:boolean')
                    elif v in ['str[]', 'string[]']:
                        props_list.append(f'{k}:string[]')
                    else:
                        props_list.append(f'{k}')

                # create list of lists and flatten
                # removes need for empty check of property list
                out_list = [[_id], props_list, [':LABEL']]
                out_list = [val for sublist in out_list for val in sublist]

                with open(header_path, 'w', encoding='utf-8') as f:

                    # concatenate with delimiter
                    row = self.delim.join(out_list)
                    f.write(row)

                # import call path for custom setup
                if self.import_call_file_prefix:
                    header_path = os.path.join(
                        self.import_call_file_prefix,
                        f'{pascal_label}-header.csv',
                    )
                    parts_path = os.path.join(
                        self.import_call_file_prefix,
                        f'{pascal_label}-part.*',
                    )

                # add file path to neo4 admin import statement
                self.import_call_nodes += (
                    f'--nodes="{header_path},{parts_path}" '
                )

        return True

    def _write_single_node_list_to_file(
        self,
        node_list: list,
        label: str,
        prop_dict: dict,
        labels: str,
    ):
        """
        This function takes one list of biocypher nodes and writes them
        to a Neo4j admin import compatible CSV file.

        Args:
            node_list (list): list of BioCypherNodes to be written
            label (str): the primary label of the node
            prop_dict (dict): properties of node class passed from parsing
                function and their types
            labels (str): string of one or several concatenated labels
                for the node class

        Returns:
            bool: The return value. True for success, False otherwise.
        """
        if not all(isinstance(n, BioCypherNode) for n in node_list):
            logger.error('Nodes must be passed as type BioCypherNode.')
            return False

        # from list of nodes to list of strings
        lines = []

        for n in node_list:

            # check for deviations in properties
            # node properties
            n_props = n.get_properties()
            n_keys = list(n_props.keys())
            # reference properties
            ref_props = list(prop_dict.keys())

            # compare lists order invariant
            if not set(ref_props) == set(n_keys):
                onode = n.get_id()
                oprop1 = set(ref_props).difference(n_keys)
                oprop2 = set(n_keys).difference(ref_props)
                logger.error(
                    f'At least one node of the class {n.get_label()} '
                    f'has more or fewer properties than another. '
                    f'Offending node: {onode!r}, offending property: '
                    f'{max([oprop1, oprop2])}. '
                    f'All reference properties: {ref_props}, '
                    f'All node properties: {n_keys}.',
                )
                return False

            line = [n.get_id()]

            if ref_props:

                plist = []
                # make all into strings, put actual strings in quotes
                for k, v in prop_dict.items():
                    p = n_props.get(k)
                    if p is None:  # TODO make field empty instead of ""?
                        plist.append('')
                    elif v in [
                        'int',
                        'long',
                        'float',
                        'double',
                        'dbl',
                        'bool',
                        'boolean',
                    ]:
                        plist.append(str(p))
                    else:
                        if isinstance(p, list):
                            plist.append(
                                self.quote + self.adelim.join(p) + self.quote
                            )
                        else:
                            plist.append(self.quote + str(p) + self.quote)

                line.append(self.delim.join(plist))
            line.append(labels)

            lines.append(self.delim.join(line) + '\n')

        # avoid writing empty files
        if lines:
            self._write_next_part(label, lines)

        return True

    def _write_edge_data(self, edges, batch_size):
        """
        Writes biocypher edges to CSV conforming to the headers created
        with `_write_edge_headers()`, and is actually required to be run
        before calling `_write_node_headers()` to set the
        :py:attr:`self.edge_property_dict` for passing the edge
        properties to the instance. Expects list or generator of edges
        from the :py:class:`BioCypherEdge` class.

        Args:
            edges (BioCypherEdge): a list or generator of edges in
                :py:class:`BioCypherEdge` format

        Returns:
            bool: The return value. True for success, False otherwise.

        Todo:
            - currently works for mixed edges but in practice often is
              called on one iterable containing one type of edge only
        """

        if isinstance(edges, GeneratorType):
            logger.debug('Writing edge CSV from generator.')

            bins = defaultdict(list)  # dict to store a list for each
            # label that is passed in
            bin_l = {}  # dict to store the length of each list for
            # batching cutoff
            reference_props = defaultdict(
                dict,
            )  # dict to store a dict of properties
            # for each label to check for consistency and their type
            # for now, relevant for `int`
            for e in edges:
                if isinstance(e, BioCypherRelAsNode):
                    # shouldn't happen any more
                    logger.error(
                        "Edges cannot be of type 'RelAsNode'. "
                        f'Caused by: {e}',
                    )
                    return False

                if not (e.get_source_id() and e.get_target_id()):
                    logger.error(
                        'Edge must have source and target node. '
                        f'Caused by: {e}',
                    )
                    continue

                label = e.get_label()

                if not label in self.seen_edges.keys():
                    self.seen_edges[label] = set()

                src_tar_id = '_'.join([e.get_source_id(), e.get_target_id()])

                # check for duplicates
                if src_tar_id in self.seen_edges.get(label, set()):
                    self.duplicate_edge_ids.add(src_tar_id)
                    if not label in self.duplicate_edge_types:
                        self.duplicate_edge_types.add(label)
                        logger.warning(
                            f'Duplicate edges found in type {label}. '
                        )
                    continue

                else:
                    self.seen_edges[label].add(src_tar_id)

                if not label in bins.keys():
                    # start new list
                    bins[label].append(e)
                    bin_l[label] = 1

                    # get properties from config if present

                    # check whether label is in ontology_adapter.leaves
                    # (may not be if it is an edge that carries the
                    # "label_as_edge" property)
                    cprops = None
                    if label in self.extended_schema:
                        cprops = self.extended_schema.get(label).get(
                            'properties',
                        )
                    else:
                        # try via "label_as_edge"
                        for k, v in self.extended_schema.items():
                            if isinstance(v, dict):
                                if v.get('label_as_edge') == label:
                                    cprops = v.get('properties')
                                    break
                    if cprops:
                        d = cprops

                        # add strict mode properties
                        if self.strict_mode:
                            d['source'] = 'str'
                            d['version'] = 'str'
                            d['licence'] = 'str'

                    else:
                        d = dict(e.get_properties())
                        # encode property type
                        for k, v in d.items():
                            if d[k] is not None:
                                d[k] = type(v).__name__
                    # else use first encountered edge to define
                    # properties for checking; could later be by
                    # checking all edges but much more complicated,
                    # particularly involving batch writing (would
                    # require "do-overs"). for now, we output a warning
                    # if edge properties diverge from reference
                    # properties (in write_single_edge_list_to_file)
                    # TODO

                    reference_props[label] = d

                else:
                    # add to list
                    bins[label].append(e)
                    bin_l[label] += 1
                    if not bin_l[label] < batch_size:
                        # batch size controlled here
                        passed = self._write_single_edge_list_to_file(
                            bins[label],
                            label,
                            reference_props[label],
                        )

                        if not passed:
                            return False

                        bins[label] = []
                        bin_l[label] = 0

            # after generator depleted, write remainder of bins
            for label, nl in bins.items():

                passed = self._write_single_edge_list_to_file(
                    nl,
                    label,
                    reference_props[label],
                )

                if not passed:
                    return False

            # use complete bin list to write header files
            # TODO if a edge type has varying properties
            # (ie missingness), we'd need to collect all possible
            # properties in the generator pass

            # save first-edge properties to instance attribute
            for label in reference_props.keys():
                self.edge_property_dict[label] = reference_props[label]

            return True
        else:
            if type(edges) is not list:
                logger.error('Edges must be passed as list or generator.')
                return False
            else:

                def gen(edges):
                    yield from edges

                return self._write_edge_data(gen(edges), batch_size=batch_size)

    def _write_edge_headers(self):
        """
        Writes single CSV file for a graph entity that is represented
        as an edge as per the definition in the `schema_config.yaml`,
        containing only the header for this type of edge.

        Returns:
            bool: The return value. True for success, False otherwise.
        """
        # load headers from data parse
        if not self.edge_property_dict:
            logger.error(
                'Header information not found. Was the data parsed first?',
            )
            return False

        for label, props in self.edge_property_dict.items():
            # create header CSV with :START_ID, (optional) properties,
            # :END_ID, :TYPE

            # translate label to PascalCase
            pascal_label = self.translator.name_sentence_to_pascal(label)

            # paths
            header_path = os.path.join(
                self.outdir,
                f'{pascal_label}-header.csv',
            )
            parts_path = os.path.join(self.outdir, f'{pascal_label}-part.*')

            # check for file exists
            if not os.path.exists(header_path):

                # concatenate key:value in props
                props_list = []
                for k, v in props.items():
                    if v in ['int', 'long']:
                        props_list.append(f'{k}:long')
                    elif v in ['float', 'double']:
                        props_list.append(f'{k}:double')
                    elif v in [
                        'bool',
                        'boolean',
                    ]:  # TODO does Neo4j support bool?
                        props_list.append(f'{k}:boolean')
                    else:
                        props_list.append(f'{k}')

                # create list of lists and flatten
                # removes need for empty check of property list
                out_list = [[':START_ID'], props_list, [':END_ID'], [':TYPE']]
                out_list = [val for sublist in out_list for val in sublist]

                with open(header_path, 'w', encoding='utf-8') as f:

                    # concatenate with delimiter
                    row = self.delim.join(out_list)
                    f.write(row)

                # import call path for custom setup
                if self.import_call_file_prefix:
                    header_path = os.path.join(
                        self.import_call_file_prefix,
                        f'{pascal_label}-header.csv',
                    )
                    parts_path = os.path.join(
                        self.import_call_file_prefix,
                        f'{pascal_label}-part.*',
                    )

                # add file path to neo4 admin import statement
                self.import_call_edges += (
                    f'--relationships="{header_path},{parts_path}" '
                )

        return True

    def _write_single_edge_list_to_file(
        self,
        edge_list: list,
        label: str,
        prop_dict: dict,
    ):
        """
        This function takes one list of biocypher edges and writes them
        to a Neo4j admin import compatible CSV file.

        Args:
            edge_list (list): list of BioCypherEdges to be written

            label (str): the label (type) of the edge

            prop_dict (dict): properties of node class passed from parsing
                function and their types

        Returns:
            bool: The return value. True for success, False otherwise.
        """

        if not all(isinstance(n, BioCypherEdge) for n in edge_list):

            logger.error('Edges must be passed as type BioCypherEdge.')
            return False

        # from list of edges to list of strings
        lines = []
        for e in edge_list:
            # check for deviations in properties
            # edge properties
            e_props = e.get_properties()
            e_keys = list(e_props.keys())
            ref_props = list(prop_dict.keys())

            # compare list order invariant
            if not set(ref_props) == set(e_keys):
                oedge = f'{e.get_source_id()}-{e.get_target_id()}'
                oprop1 = set(ref_props).difference(e_keys)
                oprop2 = set(e_keys).difference(ref_props)
                logger.error(
                    f'At least one edge of the class {e.get_label()} '
                    f'has more or fewer properties than another. '
                    f'Offending edge: {oedge!r}, offending property: '
                    f'{max([oprop1, oprop2])}. '
                    f'All reference properties: {ref_props}, '
                    f'All edge properties: {e_keys}.',
                )
                return False

            if ref_props:

                plist = []
                # make all into strings, put actual strings in quotes
                for k, v in prop_dict.items():
                    p = e_props.get(k)
                    if p is None:  # TODO make field empty instead of ""?
                        plist.append('')
                    elif v in [
                        'int',
                        'long',
                        'float',
                        'double',
                        'dbl',
                        'bool',
                        'boolean',
                    ]:
                        plist.append(str(p))
                    else:
                        if isinstance(p, list):
                            plist.append(
                                self.quote + self.adelim.join(p) + self.quote
                            )
                        elif '**' in p:
                            plist.append(
                                self.quote + p.replace('**', self.adelim) +
                                self.quote
                            )
                        else:
                            plist.append(self.quote + str(p) + self.quote)

                lines.append(
                    self.delim.join(
                        [
                            e.get_source_id(),
                            # here we need a list of properties in
                            # the same order as in the header
                            self.delim.join(plist),
                            e.get_target_id(),
                            self.translator.
                            name_sentence_to_pascal(e.get_label(), ),
                        ],
                    ) + '\n',
                )
            else:
                lines.append(
                    self.delim.join(
                        [
                            e.get_source_id(),
                            e.get_target_id(),
                            self.translator.
                            name_sentence_to_pascal(e.get_label(), ),
                        ],
                    ) + '\n',
                )

        # avoid writing empty files
        if lines:
            self._write_next_part(label, lines)

        return True

    def _write_next_part(self, label: str, lines: list):
        """
        This function writes a list of strings to a new part file.

        Args:
            label (str): the label (type) of the edge; internal
            representation sentence case -> needs to become PascalCase
            for disk representation

            lines (list): list of strings to be written

        Returns:
            bool: The return value. True for success, False otherwise.
        """
        # translate label to PascalCase
        label = self.translator.name_sentence_to_pascal(label)

        # list files in self.outdir
        files = glob.glob(os.path.join(self.outdir, f'{label}-part*.csv'))
        # find file with highest part number
        if files:
            next_part = (
                max(
                    [
                        int(
                            f.split('.')[-2].split('-')[-1].replace('part', '')
                        ) for f in files
                    ],
                ) + 1
            )
        else:
            next_part = 0

        # write to file
        padded_part = str(next_part).zfill(3)
        logger.info(
            f'Writing {len(lines)} entries to {label}-part{padded_part}.csv',
        )
        file_path = os.path.join(self.outdir, f'{label}-part{padded_part}.csv')

        with open(file_path, 'w', encoding='utf-8') as f:

            # concatenate with delimiter
            f.writelines(lines)

    def get_import_call(self) -> str:
        """
        Function to return the import call detailing folder and
        individual node and edge headers and data files, as well as
        delimiters and database name.

        Returns:
            str: a bash command for neo4j-admin import
        """

        return self._construct_import_call()

    def write_import_call(self) -> bool:
        """
        Function to write the import call detailing folder and
        individual node and edge headers and data files, as well as
        delimiters and database name, to the export folder as txt.

        Returns:
            bool: The return value. True for success, False otherwise.
        """

        file_path = os.path.join(self.outdir, 'neo4j-admin-import-call.sh')
        logger.info(f'Writing neo4j-admin import call to `{file_path}`.')

        with open(file_path, 'w', encoding='utf-8') as f:

            f.write(self._construct_import_call())

        return True

    def _construct_import_call(self) -> str:
        """
        Function to construct the import call detailing folder and
        individual node and edge headers and data files, as well as
        delimiters and database name. Built after all data has been
        processed to ensure that nodes are called before any edges.

        Returns:
            str: a bash command for neo4j-admin import
        """

        # escape backslashes in self.delim and self.adelim
        delim = self.delim.encode('unicode_escape').decode('utf-8')
        adelim = self.adelim.encode('unicode_escape').decode('utf-8')

        import_call = (
            f'{self.import_call_bin_prefix}neo4j-admin import '
            f'--database={self.db_name} '
            f'--delimiter="{delim}" --array-delimiter="{adelim}" '
        )

        if self.quote == "'":
            import_call += f'--quote="{self.quote}" '
        else:
            import_call += f"--quote='{self.quote}' "

        if self.wipe:
            import_call += f'--force=true '
        if self.skip_bad_relationships:
            import_call += '--skip-bad-relationships=true '
        if self.skip_duplicate_nodes:
            import_call += '--skip-duplicate-nodes=true '

        # append node and edge import calls
        import_call += self.import_call_nodes
        import_call += self.import_call_edges

        return import_call

    def get_duplicate_nodes(self):
        """
        Function to return a list of duplicate nodes.

        Returns:
            list: list of duplicate nodes
        """

        if self.duplicate_node_types:
            return (self.duplicate_node_types, self.duplicate_node_ids)
        else:
            return None

    def get_duplicate_edges(self):
        """
        Function to return a list of duplicate edges.

        Returns:
            list: list of duplicate edges
        """

        if self.duplicate_edge_types:
            return (self.duplicate_edge_types, self.duplicate_edge_ids)
        else:
            return None



#!/usr/bin/env python

#
# Copyright 2021, Heidelberg University Clinic
#
# File author(s): Sebastian Lobentanzer
#                 ...
#
# Distributed under MIT licence, see the file `LICENSE`.
#
"""
A wrapper around the Neo4j driver which handles the DBMS connection and
provides basic management methods.
"""
from collections.abc import Iterable

from ._logger import logger

logger.debug(f'Loading module {__name__}.')

from typing import TYPE_CHECKING, Optional
import itertools

from more_itertools import peekable

if TYPE_CHECKING:

    import neo4j

import neo4j_utils

from . import _misc
from ._write import BatchWriter
from ._config import config as _config
from ._create import VersionNode, BioCypherEdge, BioCypherNode
from ._mapping import OntologyMapping
from ._ontology import Ontology
from ._translate import Translator

__all__ = ['Driver']


class Driver(neo4j_utils.Driver):
    """
    Manages a connection to a biocypher database.

    The connection can be defined in three ways:
        * Providing a ready ``neo4j.Driver`` instance
        * By URI and authentication data
        * By a YAML config file

    Args:
        driver:
            A ``neo4j.Driver`` instance, created by, for example,
            ``neo4j.GraphDatabase.driver``.
        db_name:
            Name of the database (Neo4j graph) to use.
        db_uri:
            Protocol, host and port to access the Neo4j server.
        db_user:
            Neo4j user name.
        db_passwd:
            Password of the Neo4j user.
        fetch_size:
            Optional; the fetch size to use in database transactions.
        wipe:
            Wipe the database after connection, ensuring the data is
            loaded into an empty database.
        offline:
            Do not connect to the database, but use the provided
            schema to create a graph representation and write CSVs for
            admin import.
        output_directory:
            Directory to write CSV files to.
        increment_version:
            Whether to increase version number automatically and create a
            new BioCypher version node in the graph.
        user_schema_config_path:
            Path to the graph database schema configuration file.
        clear_cache:
            Whether to clear the ontological hierarchy cache at driver
            instantiation. The cache is used to speed up the translation
            of Biolink classes to the database schema.
        delimiter:
            Delimiter for CSV export.
        array_delimiter:
            Array delimiter for CSV exported contents.
        quote_char:
            String quotation character for CSV export.
        import_call_bin_prefix:
            Prefix for the Cypher call to the admin import shell command.
            Defaults to ``bin/``.
        import_call_file_prefix:
            Path prefix for the data files (headers and parts) in the admin
            import call. Defaults to the absolute output directory path.
        skip_bad_relationships:
            Whether to skip relationships with missing source or target
            nodes in the admin import shell command.
        skip_duplicate_nodes:
            Whether to skip duplicate nodes in the admin import shell
            command.
        biolink_model:
            Path to the custom Biolink model YAML file.
        tail_ontology_url:
            URL of the ontology to hybridise to the head ontology.
        head_join_node:
            Biolink class of the node to join the tail ontology to.
        tail_join_node:
            Ontology class of the node to join the head ontology to.
    """
    def __init__(
        self,
        driver: Optional['neo4j.Driver'] = None,
        db_name: Optional[str] = None,
        db_uri: Optional[str] = None,
        db_user: Optional[str] = None,
        db_passwd: Optional[str] = None,
        multi_db: Optional[bool] = None,
        fetch_size: int = 1000,
        skip_bad_relationships: bool = False,
        skip_duplicate_nodes: bool = False,
        wipe: bool = False,
        strict_mode: Optional[bool] = None,
        offline: Optional[bool] = None,
        output_directory: Optional[str] = None,
        increment_version: bool = True,
        clear_cache: Optional[bool] = None,
        user_schema_config_path: Optional[str] = None,
        delimiter: Optional[str] = None,
        array_delimiter: Optional[str] = None,
        quote_char: Optional[str] = None,
        import_call_bin_prefix: Optional[str] = None,
        import_call_file_prefix: Optional[str] = None,
        head_ontology: Optional[dict] = None,
        tail_ontologies: Optional[dict] = None,
    ):

        # Neo4j options
        db_name = db_name or _config('neo4j_db')
        db_uri = db_uri or _config('neo4j_uri')
        db_user = db_user or _config('neo4j_user')
        db_passwd = db_passwd or _config('neo4j_pw')
        multi_db = multi_db or _config('neo4j_multi_db')
        self.db_delim = delimiter or _config('neo4j_delimiter')
        self.db_adelim = array_delimiter or _config('neo4j_array_delimiter')
        self.db_quote = quote_char or _config('neo4j_quote_char')

        if import_call_bin_prefix is None:
            self.import_call_bin_prefix = _config(
                'neo4j_import_call_bin_prefix'
            )
        else:
            self.import_call_bin_prefix = import_call_bin_prefix

        if import_call_file_prefix is None:
            self.import_call_file_prefix = _config(
                'neo4j_import_call_file_prefix'
            )
        else:
            self.import_call_file_prefix = import_call_file_prefix

        self.skip_bad_relationships = skip_bad_relationships
        self.skip_duplicate_nodes = skip_duplicate_nodes
        self.wipe = wipe

        if offline is None:
            self._offline = _config('offline')
        else:
            self._offline = offline

        # BioCypher options
        self.user_schema_config_path = user_schema_config_path or _config(
            'user_schema_config_path'
        )
        self.strict_mode = strict_mode or _config('strict_mode')
        self.output_directory = output_directory or _config('output_directory')
        self.clear_cache = clear_cache or _config('clear_cache')

        self.head_ontology = head_ontology or _config('head_ontology')

        self.tail_ontologies = tail_ontologies or _config('tail_ontologies')

        if self._offline:

            if not self.user_schema_config_path:
                raise ValueError(
                    'Offline mode requires a user schema config file.'
                    ' Please provide one with the `user_schema_config_path`'
                    ' argument or set the `user_schema_config_path`'
                    ' configuration variable.'
                )

            logger.info('Offline mode: no connection to DBMS.')

            self._ontology_mapping = OntologyMapping(
                config_file=self.user_schema_config_path,
            )

            self._db_config = {
                'uri': db_uri,
                'user': db_user,
                'passwd': db_passwd,
                'db': db_name,
                'fetch_size': fetch_size,
            }

            self.driver = None
            self._db_name = db_name

        else:

            neo4j_utils.Driver.__init__(**locals())

            # if db representation node does not exist or explicitly
            # asked for wipe, create new graph representation: default
            # yaml, interactive?
            if wipe:

                # get database version node ('check' module) immutable
                # variable of each instance (ie, each call from the
                # adapter to BioCypher); checks for existence of graph
                # representation and returns if found, else creates new
                # one
                self._ontology_mapping = VersionNode(
                    from_config=offline or wipe,
                    config_file=self.user_schema_config_path,
                    bcy_driver=self,
                )

                # init requires db_meta to be set
                self.init_db()

            else:

                self._ontology_mapping = VersionNode(self)

        if increment_version:

            # set new current version node
            self.update_meta_graph()

        self.ontology = None
        self.batch_writer = None
        self._update_translator()

        # TODO: implement passing a driver instance
        # I am not sure, but seems like it should work from driver

    def update_meta_graph(self):

        if self._offline:
            return

        logger.info('Updating Neo4j meta graph.')
        # add version node
        self.add_biocypher_nodes(self._ontology_mapping)

        # find current version node
        db_version = self.query(
            'MATCH (v:BioCypher) '
            'WHERE NOT (v)-[:PRECEDES]->() '
            'RETURN v',
        )
        # connect version node to previous
        if db_version[0]:
            e_meta = BioCypherEdge(
                self._ontology_mapping.graph_state['id'],
                self._ontology_mapping.node_id,
                'PRECEDES',
            )
            self.add_biocypher_edges(e_meta)

        # add structure nodes
        no_l = []
        # leaves of the hierarchy specified in schema yaml
        for entity, params in self._ontology_mapping.extended_schema.items():
            no_l.append(
                BioCypherNode(
                    node_id=entity,
                    node_label='MetaNode',
                    properties=params,
                ),
            )
        self.add_biocypher_nodes(no_l)

        # remove connection of structure nodes from previous version
        # node(s)
        self.query('MATCH ()-[r:CONTAINS]-()'
                   'DELETE r', )

        # connect structure nodes to version node
        ed_v = []
        current_version = self._ontology_mapping.get_id()
        for entity in self._ontology_mapping.extended_schema.keys():
            ed_v.append(
                BioCypherEdge(
                    source_id=current_version,
                    target_id=entity,
                    relationship_label='CONTAINS',
                ),
            )
        self.add_biocypher_edges(ed_v)

        # add graph structure between MetaNodes
        ed = []
        for no in no_l:
            id = no.get_id()
            src = no.get_properties().get('source')
            tar = no.get_properties().get('target')
            if None not in [id, src, tar]:
                ed.append(BioCypherEdge(id, src, 'IS_SOURCE_OF'))
                ed.append(BioCypherEdge(id, tar, 'IS_TARGET_OF'))
        self.add_biocypher_edges(ed)

    def _update_translator(self):

        self.translator = Translator(
            extended_schema=self._ontology_mapping.extended_schema,
            strict_mode=self.strict_mode,
        )

    def init_db(self):
        """
        Used to initialise a property graph database by deleting
        contents and constraints and setting up new constraints.

        Todo:
            - set up constraint creation interactively depending on the
                need of the database
        """

        self.wipe_db()
        self._create_constraints()
        logger.info('Initialising database.')

    def _create_constraints(self):
        """
        Creates constraints on node types in the graph. Used for
        initial setup.

        Grabs leaves of the ``schema_config.yaml`` file and creates
        constraints on the id of all entities represented as nodes.
        """

        logger.info('Creating constraints for node types in config.')

        # get structure
        for leaf in self._ontology_mapping.extended_schema.items():
            label = leaf[0]
            if leaf[1]['represented_as'] == 'node':

                s = (
                    f'CREATE CONSTRAINT `{label}_id` '
                    f'IF NOT EXISTS ON (n:`{label}`) '
                    'ASSERT n.id IS UNIQUE'
                )
                self.query(s)

    def add_nodes(self, id_type_tuples: Iterable[tuple]) -> tuple:
        """
        Generic node adder method to add any kind of input to the graph via the
        :class:`biocypher.create.BioCypherNode` class. Employs translation
        functionality and calls the :meth:`add_biocypher_nodes()` method.

        Args:
            id_type_tuples (iterable of 3-tuple): for each node to add to
                the biocypher graph, a 3-tuple with the following layout:
                first, the (unique if constrained) ID of the node; second, the
                type of the node, capitalised or PascalCase and in noun form
                (Neo4j primary label, eg `:Protein`); and third, a dictionary
                of arbitrary properties the node should possess (can be empty).

        Returns:
            2-tuple: the query result of :meth:`add_biocypher_nodes()`
                - first entry: data
                - second entry: Neo4j summary.
        """

        bn = self.translator.translate_nodes(id_type_tuples)
        return self.add_biocypher_nodes(bn)

    def add_edges(self, id_src_tar_type_tuples: Iterable[tuple]) -> tuple:
        """
        Generic edge adder method to add any kind of input to the graph
        via the :class:`biocypher.create.BioCypherEdge` class. Employs
        translation functionality and calls the
        :meth:`add_biocypher_edges()` method.

        Args:

            id_src_tar_type_tuples (iterable of 5-tuple):

                for each edge to add to the biocypher graph, a 5-tuple
                with the following layout: first, the optional unique ID
                of the interaction. This can be `None` if there is no
                systematic identifier (which for many interactions is
                the case). Second and third, the (unique if constrained)
                IDs of the source and target nodes of the relationship;
                fourth, the type of the relationship; and fifth, a
                dictionary of arbitrary properties the edge should
                possess (can be empty).

        Returns:

            2-tuple: the query result of :meth:`add_biocypher_edges()`

                - first entry: data
                - second entry: Neo4j summary.
        """

        bn = self.translator.translate_edges(id_src_tar_type_tuples)
        return self.add_biocypher_edges(bn)

    def add_biocypher_nodes(
        self,
        nodes: Iterable[BioCypherNode],
        explain: bool = False,
        profile: bool = False,
    ) -> bool:
        """
        Accepts a node type handoff class
        (:class:`biocypher.create.BioCypherNode`) with id,
        label, and a dict of properties (passing on the type of
        property, ie, ``int``, ``str``, ...).

        The dict retrieved by the
        :meth:`biocypher.create.BioCypherNode.get_dict()` method is
        passed into Neo4j as a map of maps, explicitly encoding node id
        and label, and adding all other properties from the 'properties'
        key of the dict. The merge is performed via APOC, matching only
        on node id to prevent duplicates. The same properties are set on
        match and on create, irrespective of the actual event.

        Args:
            nodes:
                An iterable of :class:`biocypher.create.BioCypherNode` objects.
            explain:
                Call ``EXPLAIN`` on the CYPHER query.
            profile:
                Do profiling on the CYPHER query.

        Returns:
            True for success, False otherwise.
        """

        try:

            entities = [
                node.get_dict() for node in _misc.ensure_iterable(nodes)
            ]

        except AttributeError:

            msg = 'Nodes must have a `get_dict` method.'
            logger.error(msg)

            raise ValueError(msg)

        logger.info(f'Merging {len(entities)} nodes.')

        entity_query = (
            'UNWIND $entities AS ent '
            'CALL apoc.merge.node([ent.node_label], '
            '{id: ent.node_id}, ent.properties, ent.properties) '
            'YIELD node '
            'RETURN node'
        )

        method = 'explain' if explain else 'profile' if profile else 'query'

        result = getattr(self, method)(
            entity_query,
            parameters={
                'entities': entities,
            },
        )

        logger.info('Finished merging nodes.')

        return result

    def add_biocypher_edges(
        self,
        edges: Iterable[BioCypherEdge],
        explain: bool = False,
        profile: bool = False,
    ) -> bool:
        """
        Accepts an edge type handoff class
        (:class:`biocypher.create.BioCypherEdge`) with source
        and target ids, label, and a dict of properties (passing on the
        type of property, ie, int, string ...).

        The individual edge is either passed as a singleton, in the case
        of representation as an edge in the graph, or as a 4-tuple, in
        the case of representation as a node (with two edges connecting
        to interaction partners).

        The dict retrieved by the
        :meth:`biocypher.create.BioCypherEdge.get_dict()` method is
        passed into Neo4j as a map of maps, explicitly encoding source
        and target ids and the relationship label, and adding all edge
        properties from the 'properties' key of the dict. The merge is
        performed via APOC, matching only on source and target id to
        prevent duplicates. The same properties are set on match and on
        create, irrespective of the actual event.

        Args:
            edges:
                An iterable of :class:`biocypher.create.BioCypherEdge` objects.
            explain:
                Call ``EXPLAIN`` on the CYPHER query.
            profile:
                Do profiling on the CYPHER query.

        Returns:
            `True` for success, `False` otherwise.
        """

        edges = _misc.ensure_iterable(edges)
        edges = itertools.chain(*(_misc.ensure_iterable(i) for i in edges))

        nodes = []
        rels = []

        try:

            for e in edges:

                if hasattr(e, 'get_node'):

                    nodes.append(e.get_node())
                    rels.append(e.get_source_edge().get_dict())
                    rels.append(e.get_target_edge().get_dict())

                else:

                    rels.append(e.get_dict())

        except AttributeError:

            msg = 'Edges and nodes must have a `get_dict` method.'
            logger.error(msg)

            raise ValueError(msg)

        self.add_biocypher_nodes(nodes)
        logger.info(f'Merging {len(rels)} edges.')

        # cypher query

        # merging only on the ids of the entities, passing the
        # properties on match and on create;
        # TODO add node labels?
        node_query = (
            'UNWIND $rels AS r '
            'MERGE (src {id: r.source_id}) '
            'MERGE (tar {id: r.target_id}) '
        )

        self.query(node_query, parameters={'rels': rels})

        edge_query = (
            'UNWIND $rels AS r '
            'MATCH (src {id: r.source_id}) '
            'MATCH (tar {id: r.target_id}) '
            'WITH src, tar, r '
            'CALL apoc.merge.relationship'
            '(src, r.relationship_label, NULL, '
            'r.properties, tar, r.properties) '
            'YIELD rel '
            'RETURN rel'
        )

        method = 'explain' if explain else 'profile' if profile else 'query'

        result = getattr(self, method)(edge_query, parameters={'rels': rels})

        logger.info('Finished merging edges.')

        return result

    def write_nodes(self, nodes):
        """
        Write BioCypher nodes to disk using the :mod:`write` module,
        formatting the CSV to enable Neo4j admin import from the target
        directory.

        Args:
            nodes (iterable): collection of nodes to be written in
                BioCypher-compatible CSV format; can be any compatible
                (ie, translatable) input format or already as
                :class:`biocypher.create.BioCypherNode`.
        """

        # instantiate ontology on demand because it takes time to load
        self.start_ontology()

        self.start_batch_writer()

        nodes = peekable(nodes)
        if not isinstance(nodes.peek(), BioCypherNode):
            tnodes = self.translator.translate_nodes(nodes)
        else:
            tnodes = nodes
        # write node files
        return self.batch_writer.write_nodes(tnodes)

    def start_batch_writer(self) -> None:
        """
        Instantiate the batch writer if it does not exist.

        Args:
            dirname (str): the directory to write the files to
            db_name (str): the name of the database to write the files to
        """
        if not self.batch_writer:
            self.batch_writer = BatchWriter(
                ontology=self.ontology,
                translator=self.translator,
                delimiter=self.db_delim,
                array_delimiter=self.db_adelim,
                quote=self.db_quote,
                dirname=self.output_directory,
                db_name=self._db_name,
                skip_bad_relationships=self.skip_bad_relationships,
                skip_duplicate_nodes=self.skip_duplicate_nodes,
                import_call_bin_prefix=self.import_call_bin_prefix,
                import_call_file_prefix=self.import_call_file_prefix,
                wipe=self.wipe,
                strict_mode=self.strict_mode,
            )

    def start_ontology(self) -> None:
        """
        Instantiate the :class:`biocypher._ontology.Ontology` if not
        existing.
        """
        if not self.ontology:
            self.ontology = Ontology(
                head_ontology=self.head_ontology,
                mapping=self._ontology_mapping,
                tail_ontologies=self.tail_ontologies,
            )

    def write_edges(
        self,
        edges,
    ) -> None:
        """
        Write BioCypher edges to disk using the :mod:`write` module,
        formatting the CSV to enable Neo4j admin import from the target
        directory.

        Args:
            edges (iterable): collection of edges to be written in
                BioCypher-compatible CSV format; can be any compatible
                (ie, translatable) input format or already as
                :class:`biocypher.create.BioCypherEdge`.
        """

        # instantiate adapter on demand because it takes time to load
        # the biolink model toolkit
        self.start_ontology()

        self.start_batch_writer()

        edges = peekable(edges)
        if not isinstance(edges.peek(), BioCypherEdge):
            tedges = self.translator.translate_edges(edges)
        else:
            tedges = edges
        # write edge files
        self.batch_writer.write_edges(tedges)

    def get_import_call(self):
        """
        Upon using the batch writer for writing admin import CSV files,
        return a string containing the neo4j admin import call with
        delimiters, database name, and paths of node and edge files.

        Returns:
            str: a neo4j-admin import call
        """
        return self.batch_writer.get_import_call()

    def write_import_call(self):
        """
        Upon using the batch writer for writing admin import CSV files,
        write a string containing the neo4j admin import call with
        delimiters, database name, and paths of node and edge files, to
        the export directory.

        Returns:
            bool: The return value. True for success, False otherwise.
        """
        return self.batch_writer.write_import_call()

    def log_missing_bl_types(self):
        """
        Get the set of Biolink types encountered without an entry in
        the `schema_config.yaml` and print them to the logger.

        Returns:
            set: a set of missing Biolink types
        """

        mt = self.translator.get_missing_biolink_types()

        if mt:
            msg = (
                'Input entities not accounted for due to them not being '
                'present in the `schema_config.yaml` configuration file '
                '(this is not necessarily a problem, if you did not intend '
                'to include them in the database; see the log for details): \n'
            )
            for k, v in mt.items():
                msg += f'    {k}: {v} \n'

            logger.info(msg)
            return mt

        else:
            logger.info('No missing Biolink types in input.')
            return None

    def log_duplicates(self):
        """
        Get the set of duplicate nodes and edges encountered and print them to
        the logger.
        """

        dn = self.batch_writer.get_duplicate_nodes()

        if dn:

            ntypes = dn[0]
            nids = dn[1]

            msg = ('Duplicate node types encountered (IDs in log): \n')
            for typ in ntypes:
                msg += f'    {typ}\n'

            logger.info(msg)

            idmsg = ('Duplicate node IDs encountered: \n')
            for _id in nids:
                idmsg += f'    {_id}\n'

            logger.debug(idmsg)

        else:
            logger.info('No duplicate nodes in input.')

        de = self.batch_writer.get_duplicate_edges()

        if de:

            etypes = de[0]
            eids = de[1]

            msg = ('Duplicate edge types encountered (IDs in log): \n')
            for typ in etypes:
                msg += f'    {typ}\n'

            logger.info(msg)

            idmsg = ('Duplicate edge IDs encountered: \n')
            for _id in eids:
                idmsg += f'    {_id}\n'

            logger.debug(idmsg)

        else:
            logger.info('No duplicate edges in input.')

    def show_ontology_structure(self) -> None:
        """
        Show the ontology structure of the database using the Biolink schema and
        treelib.
        """

        self.start_ontology()

        self.ontology.show_ontology_structure()

    # TRANSLATION METHODS ###

    def translate_term(self, term: str) -> str:
        """
        Translate a term to its BioCypher equivalent.
        """

        # instantiate adapter if not exists
        self.start_ontology()

        return self.translator.translate_term(term)

    def reverse_translate_term(self, term: str) -> str:
        """
        Reverse translate a term from its BioCypher equivalent.
        """

        # instantiate adapter if not exists
        self.start_ontology()

        return self.translator.reverse_translate_term(term)

    def translate_query(self, query: str) -> str:
        """
        Translate a query to its BioCypher equivalent.
        """

        # instantiate adapter if not exists
        self.start_ontology()

        return self.translator.translate(query)

    def reverse_translate_query(self, query: str) -> str:
        """
        Reverse translate a query from its BioCypher equivalent.
        """

        # instantiate adapter if not exists
        self.start_ontology()

        return self.translator.reverse_translate(query)

    def __repr__(self):

        return f'<BioCypher {neo4j_utils.Driver.__repr__(self)[1:]}'



#!/usr/bin/env python

#
# Copyright 2021, Heidelberg University Clinic
#
# File author(s): Sebastian Lobentanzer
#                 ...
#
# Distributed under MIT licence, see the file `LICENSE`.
#
"""
Graph database standard for molecular biology
"""

__all__ = [
    '__version__',
    '__author__',
    'module_data',
    'config',
    'logfile',
    'log',
    'Driver',
]

from ._config import config, module_data
from ._driver import Driver
from ._logger import log, logfile
from ._metadata import __author__, __version__
